"""Test compare_tables.py."""

import pytest
import pandas as pd
from pandas.testing import assert_frame_equal

from tabled.compare_tables import dataframe_diffs, InvalidComparison


def test_dataframe_diffs():
    # Test case 1: Different columns
    df1 = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=[0, 1])
    df2 = pd.DataFrame({"A": [1, 2], "C": [5, 6]}, index=[0, 1])
    diffs = dataframe_diffs(df1, df2)
    assert "columns_diff" in diffs
    assert diffs["columns_diff"] == {"left_right": {"B"}, "right_left": {"C"}}

    # Test case 2: Different indices
    df1 = pd.DataFrame({"A": [1, 2]}, index=[0, 1])
    df2 = pd.DataFrame({"A": [1, 2]}, index=[1, 2])
    diffs = dataframe_diffs(df1, df2)
    assert "index_diff" in diffs
    assert diffs["index_diff"] == {"left_right": {0}, "right_left": {2}}

    # Test case 3: Different shapes
    df1 = pd.DataFrame({"A": [1, 2, 3]}, index=[0, 1, 2])
    df2 = pd.DataFrame({"A": [1, 2]}, index=[0, 1])
    diffs = dataframe_diffs(df1, df2)
    assert "shape_diff" in diffs
    assert diffs["shape_diff"] == ((3, 1), (2, 1))

    # Test case 4: Different data types
    df1 = pd.DataFrame({"A": [1, 2]}, index=[0, 1])
    df2 = pd.DataFrame({"A": [1.0, 2.0]}, index=[0, 1])
    diffs = dataframe_diffs(df1, df2)
    assert "dtypes_diff" in diffs
    assert diffs["dtypes_diff"] == {"A": (df1["A"].dtype, df2["A"].dtype)}

    # Test case 5: Different values
    df1 = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=[0, 1])
    df2 = pd.DataFrame({"A": [1, 5], "B": [3, 4]}, index=[0, 1])
    diffs = dataframe_diffs(df1, df2)
    assert "columns_value_diff" in diffs
    assert "A" in diffs["columns_value_diff"]
    expected_diff = pd.DataFrame({"left": [2], "right": [5]}, index=[1], dtype=float)
    assert_frame_equal(
        diffs["columns_value_diff"]["A"], expected_diff, check_dtype=False
    )

    # Test case 6: No differences
    df1 = pd.DataFrame({"A": [1, 2]}, index=[0, 1])
    df2 = pd.DataFrame({"A": [1, 2]}, index=[0, 1])
    diffs = dataframe_diffs(df1, df2)
    assert diffs == {}

    # Test case 7: Non-unique indices
    df1 = pd.DataFrame({"A": [1, 2, 3]}, index=[0, 0, 1])
    df2 = pd.DataFrame({"A": [1, 2, 3]}, index=[0, 1, 1])
    diffs = dataframe_diffs(df1, df2)
    assert isinstance(diffs["columns_value_diff"], InvalidComparison)

    # Test case 8: Custom comparison functions
    def custom_comparison(df1, df2):
        # Note we use != instead of == because we want to return True if the DataFrames are different
        # (By default, comparison values that resolve to False are not included in the output)
        return not df1.equals(df2)

    diffs = dataframe_diffs(df1, df2, comparisons=[custom_comparison])
    assert "custom_comparison" in diffs
    assert diffs["custom_comparison"]

    # Test case 9: Using diff_condition to filter results
    diffs = dataframe_diffs(df1, df2, diff_condition=lambda x: True)
    # All comparisons should be included regardless of their result
    assert "columns_diff" in diffs
    assert "index_diff" in diffs
    assert "shape_diff" in diffs
    assert "columns_value_diff" in diffs
    assert "dtypes_diff" in diffs

    # Test case 10: DataFrames with NaN values
    df1 = pd.DataFrame({"A": [1, 2, None]}, index=[0, 1, 2])
    df2 = pd.DataFrame({"A": [1, None, 3]}, index=[0, 1, 2])
    diffs = dataframe_diffs(df1, df2)
    assert "columns_value_diff" in diffs
    expected_diff = pd.DataFrame(
        {"left": [2.0, None], "right": [None, 3.0]}, index=[1, 2]
    )
    assert_frame_equal(
        diffs["columns_value_diff"]["A"], expected_diff, check_dtype=False
    )


test_dataframe_diffs()



"""
A  (key-value) data-object-layer to get (pandas) tables from a variety of sources with ease.

"""

from tabled.html import get_tables_from_url
from tabled.base import (
    get_table,  # get a table from a variety of sources
    DfFiles,  # a store (mapping from keys to dataframes)
    DfLocalFileReader,  # will be deprecated
    KeyFuncReader,
    DataframeKvReader,
    dflt_ext_mapping,
    df_from_data_according_to_key,
    file_extension,
    # get_file_ext,
    identity,
    key_func_mapping,
    split_keys,
)
from tabled.multi import ColumnOrientedMapping
from tabled.util import (
    collapse_rows,  # collapse rows in a dataframe
    expand_rows,  # expand rows in a dataframe
    collapse_columns,  # collapse columns in a dataframe
    expand_columns,  # expand columns in a dataframe
    auto_decode_bytes,  # Decode a byte sequence into a string, trying charset_normalizer gueses if fails.
    PandasJSONEncoder,  # a json encoder that can handle pandas and numpy objects
    pandas_json_dumps,  # dump a pandas object to a json string
    ensure_columns,  # ensure that the columns are in the dataframe
    ensure_first_columns,  # ensure that the columns are the first columns in the dataframe
    ensure_last_columns,  # ensure that the columns are the last columns in the dataframe
)
from tabled.misc import (
    scalar_columns,  # get list of columns that are scalar (therefore serializable to CSV)
)
from tabled.wrappers import (
    extension_to_encoder,  # get an encoder for a given extension
    extension_to_decoder,  # get a decoder for a given extension
    extension_based_decoding,  # decode a value based on the extension of the key
    extension_based_encoding,  # encode a value based on the extension of the key
    extension_based_wrap,  # Add extension-based encoding and decoding to a store
)
from tabled.compare_tables import (
    dataframe_diffs,  # compare two dataframes using specified comparison functions
)



import pandas as pd
import pytest
from tabled.join_tables import *


@pytest.fixture
def simple_tables():
    return {
        "A": pd.DataFrame({"x": [1, 2], "y": [3, 4]}),
        "B": pd.DataFrame({"y": [3, 4], "z": [5, 6]}),
    }


@pytest.fixture
def complex_tables():
    return {
        "A": pd.DataFrame({"a": [1, 2], "b": [3, 4], "c": [5, 6]}),
        "B": pd.DataFrame({"b": [3, 4], "d": [7, 8]}),
        "C": pd.DataFrame({"d": [7, 8], "e": [9, 10], "f": [11, 12]}),
        "D": pd.DataFrame({"f": [11, 12], "g": [13, 14]}),
    }


@pytest.fixture
def tables():
    return {
        "A": pd.DataFrame({"b": [1, 2, 3, 33], "c": [4, 5, 6, 66]}),
        "B": pd.DataFrame(
            {
                "b": [1, 2, 3],
                "a": [4, 5, 6],
                "d": [7, 8, 9],
                "e": [10, 11, 12],
                "f": [13, 14, 15],
            }
        ),
        "C": pd.DataFrame({"f": [13, 14, 15], "g": [4, 5, 6]}),
        "D": pd.DataFrame(
            {"d": [7, 8, 77], "e": [10, 11, 77], "h": [7, 8, 9], "i": [1, 2, 3]}
        ),
        "E": pd.DataFrame({"i": [1, 2, 3], "j": [4, 5, 6]}),
    }


# Test for the simple case
def test_minimum_covering_tree_simple(simple_tables):
    target_subset = ["y", "z"]
    edges = minimum_covering_tree(simple_tables, target_subset)
    assert set(edges) == {
        ("A", "B")
    }, f"Expected edges: {set([('A', 'B')])}, but got: {set(edges)}"


def test_generate_join_sequence_simple(simple_tables):
    target_subset = {"y", "z"}
    leaf_edges = get_leaf_edges(simple_tables, target_subset)
    join_sequence = generate_join_sequence(simple_tables, leaf_edges, target_subset)
    expected_sequence = ["A", Join("B", remove=["x"])]
    assert (
        join_sequence == expected_sequence
    ), f"Expected join sequence: {expected_sequence}, but got: {join_sequence}"


def test_compute_join_resolution_simple(simple_tables):
    target_subset = {"y", "z"}
    leaf_edges = get_leaf_edges(simple_tables, target_subset)
    join_sequence = generate_join_sequence(simple_tables, leaf_edges, target_subset)
    expected_result = pd.DataFrame({"y": [3, 4], "z": [5, 6]})
    result = compute_join_resolution(join_sequence, simple_tables)
    pd.testing.assert_frame_equal(result, expected_result)


# Test for the complex case
# def test_minimum_covering_tree_complex(complex_tables):
#     target_subset = ['a', 'e', 'g']
#     edges = minimum_covering_tree(complex_tables, target_subset)
#     expected_edges = {('A', 'B'), ('B', 'C'), ('C', 'D')}
#     assert set(edges) == expected_edges, f"Expected edges: {expected_edges}, but got: {set(edges)}"

# def test_generate_join_sequence_complex(complex_tables):
#     target_subset = {'a', 'e', 'g'}
#     leaf_edges = get_leaf_edges(complex_tables, target_subset)
#     join_sequence = generate_join_sequence(complex_tables, leaf_edges, target_subset)
#     expected_sequence = ['A', Join('B', remove=['c']), Join('C', remove=['b']), Join('D', remove=['d', 'f'])]
#     print(join_sequence)
#     assert join_sequence == expected_sequence, f"Expected join sequence: {expected_sequence}, but got: {join_sequence}"

# def test_compute_join_resolution_complex(complex_tables):
#     target_subset = {'a', 'e', 'g'}
#     leaf_edges = get_leaf_edges(complex_tables,target_subset)
#     join_sequence = generate_join_sequence(complex_tables, leaf_edges, target_subset)
#     expected_result = pd.DataFrame({'a': [1, 2], 'e': [9, 10], 'g': [13, 14]})
#     result = compute_join_resolution(join_sequence, complex_tables)
#     pd.testing.assert_frame_equal(result, expected_result)


# Test for tables
def test_minimum_covering_tree(tables):
    target_subset = {"b", "g", "j"}
    edges = minimum_covering_tree(tables, target_subset)
    expected_edges = {("A", "B"), ("B", "C"), ("B", "D"), ("D", "E")}
    assert (
        set(edges) == expected_edges
    ), f"Expected edges: {expected_edges}, but got: {set(edges)}"


def test_generate_join_sequence(tables):
    target_subset = {"b", "g", "j"}
    leaf_edges = get_leaf_edges(tables, target_subset)
    join_sequence = generate_join_sequence(tables, leaf_edges, target_subset)
    expected_sequence = [
        "B",
        Join("C", remove=["f", "a"]),
        Join("D", remove=["e", "d", "h"]),
        Join("E", remove=["i"]),
    ]
    assert (
        join_sequence == expected_sequence
    ), f"Expected join sequence: {expected_sequence}, but got: {join_sequence}"


def test_compute_join_resolution(tables):
    target_subset = {"b", "g", "j"}
    leaf_edges = get_leaf_edges(tables, target_subset)
    join_sequence = generate_join_sequence(tables, leaf_edges, target_subset)
    expected_result = pd.DataFrame({"b": [1, 2], "g": [4, 5], "j": [4, 5]})
    result = compute_join_resolution(join_sequence, tables)
    pd.testing.assert_frame_equal(result, expected_result)


from tabled.multi import execute_commands
from tabled.multi import execute_table_commands


def test_execute_commands_simply():
    from tabled.multi import Join, Remove, Load, Rename

    # ---------------------------------------------
    # First silly test

    silly_interpreter_map = {
        Join: lambda scope, command: f"Joining {command.table_key}",
        Remove: lambda scope, command: f"Removing {command.fields}",
    }

    g = execute_commands(
        [Join("asdf"), Remove("apple")], scope={}, interpreter_map=silly_interpreter_map
    )
    assert list(g) == ["Joining asdf", "Removing apple"]

    # ---------------------------------------------
    # The real case

    table1 = pd.DataFrame({"ID": [1, 2, 3], "Name": ["Alice", "Bob", "Charlie"]})
    table2 = pd.DataFrame({"ID": [2, 3, 4], "Age": [25, 30, 22]})
    table3 = pd.DataFrame({"ID": [1, 2, 3, 4], "Salary": [50000, 60000, 70000, 55000]})

    tables = {"table1": table1, "table2": table2, "table3": table3}

    commands = [
        Load("table1"),
        Rename({"Name": "First Name"}),
        Remove(["First Name"]),
        Join("table3"),
    ]

    scope = tables
    extra_scope = dict()

    it = execute_table_commands(commands, tables, extra_scope=extra_scope)

    def are_equal(a, b):
        if isinstance(a, pd.DataFrame) and isinstance(b, pd.DataFrame):
            return (a == b).all().all()
        else:
            return a == b

    next(it)
    assert are_equal(extra_scope["cumul"], scope["table1"])
    next(it)
    assert list(extra_scope["cumul"]) == ["ID", "First Name"]
    next(it)
    assert are_equal(extra_scope["cumul"], pd.DataFrame({"ID": [1, 2, 3]}))
    next(it)
    assert list(extra_scope["cumul"]) == ["ID", "Salary"]


# Test wiki table
from tabled.html import *


def test_extract_wikipedia_tables():
    wikiurl = "https://fr.wikipedia.org/wiki/Liste_des_communes_de_France_les_plus_peupl%C3%A9es"
    converter = url_to_html_func("requests")
    tables = get_tables_from_url(wikiurl, url_to_html=converter)
    assert tables is not None
    assert len(tables) > 0
    for idx, df in enumerate(tables):
        assert isinstance(df, pd.DataFrame)
        assert not df.empty


@pytest.fixture
def extracted_dataframes():
    converter = url_to_html_func("requests")
    url_aeroports_frequentes = "https://fr.wikipedia.org/wiki/Liste_des_a%C3%A9roports_les_plus_fr%C3%A9quent%C3%A9s_en_France"
    url_aeroports_vastes = "https://fr.wikipedia.org/wiki/Liste_des_a%C3%A9roports_les_plus_vastes_au_monde"

    dfs_aeroports_frequentes = get_tables_from_url(
        url_aeroports_frequentes, url_to_html=converter
    )
    dfs_aeroports_vastes = get_tables_from_url(
        url_aeroports_vastes, url_to_html=converter
    )

    return dfs_aeroports_frequentes, dfs_aeroports_vastes


def test_execute_commands_wiki(extracted_dataframes):
    from tabled.multi import Join, Remove, Load, Rename

    table1_wiki = extracted_dataframes[0][0]
    table2_wiki = extracted_dataframes[1][0]

    tables = {"table1_wiki": table1_wiki, "table2_wiki": table2_wiki}
    commands = [
        Load("table2_wiki"),
        Remove("Aéroport"),
        Rename({"Code": "Code IATA"}),
        Join("table1_wiki"),
    ]

    scope = tables
    extra_scope = dict()
    it = execute_table_commands(commands, scope, extra_scope=extra_scope)
    next(it)
    next(it)
    next(it)
    next(it)
    assert extra_scope["cumul"].shape[0] == 1



"""
This module provides functionality for joining multiple tables (pandas DataFrames) based on a target subset of columns.
It includes classes and functions to determine the optimal sequence of joins and the fields to remove during the join process.
Classes:
    Join: Represents a join operation with optional fields to remove.
Functions:
    minimum_covering_tree(tables, target_subset, start_node=None):
        Computes the minimum covering tree for the given tables and target subset of columns.
    get_leaf_edges(tables, target_subset, start_node=None):
        Retrieves the leaf edges of the minimum covering tree for the given tables and target subset of columns.
    update_leaf_edges_after_removal(tables, target_sub_set, current_leaf_edges):
        Updates the list of leaf edges after removing an edge, ensuring that the resulting leaf edges do not lead to the loss of any elements in the target subset.
    determine_remove_fields(labeled_sets, target_sub_set, joined_tables, current_table):
        Determines which fields should be removed for a given table to ensure the target subset remains covered.
    generate_join_sequence(tables, leaf_edges, target_sub_set):
        Generates a sequence of joins with remove commands based on leaf edges and the target subset of columns.
    ensure_join_op(obj):
        Ensures that the given object is a Join instance.
    compute_join_resolution(resolution_sequence, tables):
        Carries out the join operations specified in the resolution sequence with the given tables.

Example:
    >>> tables = {
    ...     'A': pd.DataFrame({'b': [1, 2, 3, 33], 'c': [4, 5, 6, 66]}),
    ...     'B': pd.DataFrame(
    ...         {
    ...             'b': [1, 2, 3],
    ...             'a': [4, 5, 6],
    ...             'd': [7, 8, 9],
    ...             'e': [10, 11, 12],
    ...             'f': [13, 14, 15],
    ...         }
    ...     ),
    ...     'C': pd.DataFrame({'f': [13, 14, 15], 'g': [4, 5, 6]}),
    ...     'D': pd.DataFrame(
    ...         {'d': [7, 8, 77], 'e': [10, 11, 77], 'h': [7, 8, 9], 'i': [1, 2, 3]}
    ...     ),
    ...     'E': pd.DataFrame({'i': [1, 2, 3], 'j': [4, 5, 6]}),
    ... }
    >>> target_sub_set = {'b', 'g', 'j'}
    >>> leaf_edges = get_leaf_edges(tables, target_sub_set)
    >>> leaf_edges
    [('B', 'C'), ('D', 'E')]
    >>> join_sequence = generate_join_sequence(tables, leaf_edges, target_sub_set)
    >>> join_sequence
    ['B', Join('C', remove=['a', 'f']), Join('D', remove=['d', 'e', 'h']), Join('E', remove=['i'])]
    >>> join_result = compute_join_resolution(join_sequence, tables)
    >>> join_result
       b  g  j
    0  1  4  4
    1  2  5  5

"""

from typing import KT, VT, Literal, Iterable, Callable, List, Tuple, Set, Mapping, Dict
import pandas as pd
from collections import deque
from tabled.util import (
    intersection_graph,
    invert_labeled_collection,
    map_values,
    breadth_first_traversal,
)


class Join:
    def __init__(self, table_id: str, remove: List[str] = None):
        self.table_id = table_id
        self.remove = remove or []

    def __repr__(self):
        # note: sorting self.remove so doctests are consistent
        remove_str = f", remove={sorted(self.remove)}" if self.remove else ""
        return f"Join('{self.table_id}'{remove_str})"

    def __eq__(self, other):
        if not isinstance(other, Join):
            return False
        return self.table_id == other.table_id and set(self.remove) == set(other.remove)


def minimum_covering_tree(
    tables: Mapping[str, pd.DataFrame],
    target_subset: Iterable[VT],
    start_node=None,
):

    labeled_sets = {table_id: set(df.columns) for table_id, df in tables.items()}
    intersections = intersection_graph(labeled_sets, edge_labels="elements")
    graph = {k: list(v) for k, v in intersections.items()}
    if start_node is None:  # if not start_node is given
        start_node = next(iter(graph))  # take the first node
    target_sub_set = set(target_subset)
    traversal = breadth_first_traversal(graph, start_node, yield_edges=True)

    edges = list()
    covered = set()

    for set_label_1, set_label_2 in traversal:
        if target_sub_set.issubset(covered):
            break
        covered |= labeled_sets[set_label_1] | labeled_sets[set_label_2]
        edges.append((set_label_1, set_label_2))

    return edges


def get_leaf_edges(
    tables: Mapping[str, pd.DataFrame], target_subset: Set[VT], start_node: KT = None
) -> List[Tuple[KT, KT]]:
    labeled_sets = {table_id: set(df.columns) for table_id, df in tables.items()}
    intersections = intersection_graph(labeled_sets, edge_labels="elements")
    graph = {k: list(v) for k, v in intersections.items()}
    if start_node is None:  # if no start_node is given
        start_node = next(iter(graph))  # take the first node

    traversal = breadth_first_traversal(graph, start_node, yield_edges=True)

    edges = []  # List to store edges of the covering tree
    covered = set()  # Set to store covered elements
    node_connections = {}  # Dictionary to track the number of connections for each node

    for node1, node2 in traversal:
        if target_subset.issubset(covered):
            break

        # Update covered elements and the number of connections for each node
        covered |= labeled_sets[node1] | labeled_sets[node2]
        node_connections[node1] = node_connections.get(node1, 0) + 1
        node_connections[node2] = node_connections.get(node2, 0) + 1
        edges.append((node1, node2))

    # Filter edges to return only those that lead to a node with one connection
    leaf_edges = [edge for edge in edges if node_connections[edge[1]] == 1]

    return leaf_edges


def update_leaf_edges_after_removal(
    tables: Mapping[str, pd.DataFrame],
    target_sub_set: Set[str],
    current_leaf_edges: List[Tuple[str, str]],
) -> List[Tuple[str, str]]:
    """
    Update the list of leaf edges after removing an edge, ensuring that the resulting
    leaf edges do not lead to the loss of any elements in the target subset.

    :param graph: The graph represented as an adjacency list.
    :param labeled_sets: The sets of elements labeled by nodes.
    :param target_subset: The target subset of elements that must remain covered.
    :param current_leaf_edges: The current list of leaf edges.
    :return: An updated list of leaf edges.
    """
    new_leaf_edges = []
    labeled_sets = {table_id: set(df.columns) for table_id, df in tables.items()}
    intersections = intersection_graph(labeled_sets, edge_labels="elements")
    graph = {k: list(v) for k, v in intersections.items()}
    for node1, node2 in current_leaf_edges:
        # Use symmetric difference to consider the edge as undirected
        unique_elements = (
            (labeled_sets[node1] - labeled_sets[node2])
            | (labeled_sets[node2] - labeled_sets[node1])
        ) & target_sub_set

        # Check if the unique elements do not include any from the target subset
        if not unique_elements:
            # Remove the edge from the graph
            graph[node1].remove(node2)
            graph[node2].remove(node1)

        else:
            # If the edge cannot be removed, add it to the new list of leaf edges
            new_leaf_edges.append((node1, node2))

    # Recalculate the leaf edges after removals
    for node, neighbors in graph.items():
        if len(neighbors) == 1:  # If the node is now a leaf node
            neighbor = neighbors[0]
            if (node, neighbor) not in new_leaf_edges and (
                neighbor,
                node,
            ) not in new_leaf_edges:
                new_leaf_edges.append((node, neighbor))

    return new_leaf_edges


def determine_remove_fields(
    labeled_sets: Dict[str, Set[str]],
    target_sub_set: Set[str],
    joined_tables: Set[str],
    current_table: str,
) -> List[str]:
    """
    Determine which fields should be removed for a given table

    :param labeled_sets: The sets of elements labeled by nodes
    :param target_sub_set: The target subset of elements that must remain covered
    :param joined_tables: The set of tables that have been or will be joined
    :param current_table: The current table being processed
    :return: A list of fields to remove
    """
    fields = labeled_sets[current_table]
    removable_fields = [
        field
        for field in fields
        if field not in target_sub_set
        and all(
            field not in labeled_sets[table]
            for table in joined_tables
            if table != current_table
        )
    ]
    return removable_fields


def generate_join_sequence(
    tables: Mapping[str, pd.DataFrame],
    leaf_edges: List[Tuple[str, str]],
    target_sub_set: Set[str],
) -> List[Join]:
    """
    Generate a sequence of joins with remove commands based on leaf edges

    :param leaf_edges: The list of leaf edges to process
    :param labeled_sets: The sets of elements labeled by nodes
    :param target_sub_set: The target subset of elements that must remain covered
    :return: A list of Join operations
    """
    labeled_sets = {table_id: set(df.columns) for table_id, df in tables.items()}
    joins = []  # Stores the sequence of joins
    joined_tables = set()  # Tracks tables that have been joined
    future_joins = {
        node for edge in leaf_edges for node in edge
    }  # Tracks all future joins

    # Start with the first node of the first edge without removals
    first_node = leaf_edges[0][0]
    first_node_remove_fields = determine_remove_fields(
        labeled_sets, target_sub_set, future_joins, first_node
    )
    joins.append(first_node)
    for node1, node2 in leaf_edges:
        for node in (node1, node2):
            if node not in joined_tables:
                # Determine the fields to remove considering future joins
                remove_fields = determine_remove_fields(
                    labeled_sets, target_sub_set, future_joins, node
                )
                if node == first_node:
                    remove_fields = first_node_remove_fields
                elif (
                    node in leaf_edges[0]
                ):  # If the node is to join with the first node
                    remove_fields.extend(
                        first_node_remove_fields
                    )  # Append remove_fields from the first node

                if node != first_node:
                    joins.append(Join(node, remove_fields))
                # Update the tracking sets
                joined_tables.add(node)
                future_joins.remove(node)

    return joins


def ensure_join_op(obj):
    if not isinstance(obj, Join):
        return Join(obj)
    return obj


def compute_join_resolution(
    resolution_sequence: Iterable, tables: Mapping[str, pd.DataFrame]
) -> pd.DataFrame:
    """
    Carries `resolution_sequence` join operations out with tables taken from `tables`.

    :param resolution_sequence: An iterable of join operations to carry out.
        Each join operation is either a table name (str) or a Join object.
        If it's a Join object, it's assumed that the table has already been joined
        and the fields to remove are in the `remove` attribute of the object.
    :param tables: A mapping of table names to tables (pd.DataFrame)
    """
    join_ops = map(ensure_join_op, resolution_sequence)
    table_key = next(join_ops).table_id
    joined = tables[table_key]
    for join_op in join_ops:
        table = tables[join_op.table_id]
        joined = joined.merge(table, how="inner")
        if join_op.remove:
            remove_cols = set(join_op.remove) & set(joined.columns)
            joined = joined.drop(columns=remove_cols)
    return joined


if __name__ == "__main__":

    tables = {
        "A": pd.DataFrame({"b": [1, 2, 3, 33], "c": [4, 5, 6, 66]}),
        "B": pd.DataFrame(
            {
                "b": [1, 2, 3],
                "a": [4, 5, 6],
                "d": [7, 8, 9],
                "e": [10, 11, 12],
                "f": [13, 14, 15],
            }
        ),
        "C": pd.DataFrame({"f": [13, 14, 15], "g": [4, 5, 6]}),
        "D": pd.DataFrame(
            {"d": [7, 8, 77], "e": [10, 11, 77], "h": [7, 8, 9], "i": [1, 2, 3]}
        ),
        "E": pd.DataFrame({"i": [1, 2, 3], "j": [4, 5, 6]}),
    }
    target_sub_set = {"b", "g", "j"}

    leaf_edges = get_leaf_edges(tables, target_sub_set)
    print(leaf_edges)
    join_sequence = generate_join_sequence(tables, leaf_edges, target_sub_set)
    print(join_sequence)
    join_result = compute_join_resolution(join_sequence, tables)
    print(join_result)

    # expected_join_resolution = [
    #     'B',
    #     Join('C', remove=['a', 'f']),
    #     Join('D', remove=['d', 'e', 'h']),
    #     Join('E', remove=['i'])
    # ]
    # expected_result = pd.DataFrame({
    #     'b': [1, 2],
    #     'g': [4, 5],
    #     'j': [4, 5]
    # })



"""Tools to compare tables"""

from typing import Callable, Sequence, Union, Any
import pandas as pd

ComparisonName = str
ComparisonFunc = Callable[[pd.DataFrame, pd.DataFrame], Any]
ComparisonsDict = dict[ComparisonName, ComparisonFunc]
ComparisonsList = Sequence[Union[ComparisonName, ComparisonFunc]]
Comparisons = Union[ComparisonsDict, ComparisonsList, ComparisonName, ComparisonFunc]


class BinaryFuncResult(dict):
    def __bool__(self):
        return any(self.values())

    @classmethod
    def from_func(cls, func, x, y):
        return cls(
            left_right=func(x, y),
            right_left=func(y, x),
        )


def columns_diff(df1: pd.DataFrame, df2: pd.DataFrame):
    """Return columns that are not common between df1 and df2."""
    return BinaryFuncResult.from_func(
        lambda x, y: x - y, set(df1.columns), set(df2.columns)
    )


def index_diff(df1: pd.DataFrame, df2: pd.DataFrame):
    """Return indices that are not common between df1 and df2."""
    return BinaryFuncResult.from_func(
        lambda x, y: x - y, set(df1.index), set(df2.index)
    )


def shape_diff(df1: pd.DataFrame, df2: pd.DataFrame):
    """Return the shapes of df1 and df2 if they differ."""
    df1_shape = df1.shape
    df2_shape = df2.shape
    if df1_shape != df2_shape:
        return df1_shape, df2_shape


def columns_value_diff(df1: pd.DataFrame, df2: pd.DataFrame):
    """
    For each column present in both dataframes, compare the values row-wise.
    Returns a dict with column names as keys and DataFrames of differences as values.
    Only columns with differences are included.
    """
    if not df1.index.is_unique or not df2.index.is_unique:
        return InvalidComparison("Indices are not unique.")
    common_index = df1.index.intersection(df2.index)
    common_columns = df1.columns.intersection(df2.columns)
    diffs = {}
    for col in common_columns:
        s1 = df1.loc[common_index, col]
        s2 = df2.loc[common_index, col]
        diff_s = s1.compare(s2, keep_shape=False, keep_equal=False)
        if not diff_s.empty:
            diff_s = diff_s.rename(columns={"self": "left", "other": "right"})
            diffs[col] = diff_s
    return diffs


def dtypes_diff(df1: pd.DataFrame, df2: pd.DataFrame):
    """Return columns where the data types differ between df1 and df2."""
    common_columns = df1.columns.intersection(df2.columns)
    diffs = {}
    for col in common_columns:
        dtype1 = df1[col].dtype
        dtype2 = df2[col].dtype
        if dtype1 != dtype2:
            diffs[col] = (dtype1, dtype2)
    return diffs


class InvalidComparison(ValueError):
    """Raised or returned when a comparison that was asked for is not applicable to the dataframes in question"""


DFLT_COMPARISONS: ComparisonsDict = {
    "columns_diff": columns_diff,
    "index_diff": index_diff,
    "shape_diff": shape_diff,
    "columns_value_diff": columns_value_diff,
    "dtypes_diff": dtypes_diff,
    # ... insert more here
}


def ensure_comparisons_dict(comparisons: Comparisons) -> ComparisonsDict:
    """Ensure that the comparisons are in the form of a dictionary."""
    if isinstance(comparisons, dict):
        return comparisons
    elif isinstance(comparisons, (str, Callable)):
        comparisons = [comparisons]

    def comparison_name_and_func(comparisons):
        for comparison in comparisons:
            if isinstance(comparison, str):
                yield comparison, DFLT_COMPARISONS[comparison]
            else:
                yield comparison.__name__, comparison

    return dict(comparison_name_and_func(comparisons))


def dataframe_diffs(
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    comparisons: Comparisons = DFLT_COMPARISONS,
    *,
    diff_condition=bool
) -> dict:
    """
    Compare the diff of dataframes using specified diff comparison functions.

    Returns a dictionary with comparison names as keys and comparison results as values.

    Parameters:
        df1 (pd.DataFrame): The first dataframe to compare.
        df2 (pd.DataFrame): The second dataframe to compare.
        comparisons (Comparisons): A dictionary or list of comparison functions or names.
            Defaults to DFLT_COMPARISONS.
        diff_condition (callable): A function that determines whether to include a comparison result
            in the output dictionary based on the comparison result.
            Defaults to the built-in `bool` function.

    Returns:
        dict: A dictionary with comparison names as keys and comparison results as values.

    Example:
        >>> import pandas as pd
        >>> df1 = pd.DataFrame({'A': [1, 2], 'B': [3, 4]}, index=[0, 1])
        >>> df2 = pd.DataFrame({'A': [1, 2], 'C': [5, 6]}, index=[1, 2])
        >>> diffs = dataframe_diffs(df1, df2)
        >>> diffs  # doctest: +NORMALIZE_WHITESPACE
        {'columns_diff': {'left_right': {'B'}, 'right_left': {'C'}},
         'index_diff': {'left_right': {0}, 'right_left': {2}},
         'columns_value_diff': {'A':    left  right
        1     2      1}}
    """
    comparisons = ensure_comparisons_dict(comparisons)
    diffs = {}
    for name, comparison in comparisons.items():
        diff_value = comparison(df1, df2)
        if diff_condition(diff_value):
            diffs[name] = diff_value
    return diffs



"""A module containing datas that will be used for tests"""

import pandas as pd

# Define the data as dictionaries
misc_small_dicts = {
    "pet_olympics": {
        "pet_name": ["Fluffy", "Barkley", "Chirpy", "Hoppy"],
        "species": ["Cat", "Dog", "Bird", "Rabbit"],
        "event": ["High Jump", "Fetch", "Singing", "Long Jump"],
        "score": [8.5, 9.2, 7.8, 8.8],
        "qualified": [True, True, False, True],
    },
    "alien_coffee_preferences": {
        "alien_name": ["Zork", "Blip", "Klorg", "Mleeb"],
        "planet": ["Zeta-2", "Plorkon", "Glibnar", "Earth"],
        "prefers_caffeine": [True, False, True, True],
        "favorite_temp": [42.0, 273.15, 37.5, 60.0],
        "comment": [
            "'Extra strong!'",
            "'I prefer slush.'",
            "'Room temp is fine.'",
            "'Hot, please!'",
        ],
    },
    "superhero_laundry_day": {
        "name": ["Superman", "Batman", "Spider-Man"],
        "power_level": [1000, 700, 500],
        "has_clean_cape": [True, False, True],
        "favorite_detergent": ["Kryptonite-Free", "Bat-OxyBoost", "StickyClean"],
    },
    "fantasy_tavern_menu": {
        "item": ["Dragon Ale", "Elf Bread", "Goblin Stew"],
        "price": [7.5, 3.0, 5.5],
        "is_alcoholic": [True, False, False],
        "servings_left": [12, 25, 8],
    },
    "alien_abduction_log": {
        "abductee_name": ["Bob", "Alice", "Zork"],
        "location": ["Kansas City", "Roswell", "Jupiter"],
        "duration_minutes": [15, 120, 30],
        "was_returned": [True, False, True],
    },
    "cat_fashion_show": {
        "cat_name": ["Mr. Whiskers", "Mittens", "Fluffy"],
        "outfit_color": ["Pink", "Blue", "Green"],
        "has_hat": [True, False, True],
        "sassiness_level": [9.5, 7.0, 8.2],
    },
    "robot_fight_club": {
        "robot_name": ["RoboX", "MechaZilla", "AIronMan"],
        "weapon": ["Laser Sword", "Rocket Fist", "Nano Dagger"],
        "power_level": [150, 180, 140],
    },
}

# Create DataFrames from the dictionaries
misc_small_dfs = {k: pd.DataFrame(v) for k, v in misc_small_dicts.items()}

superhero_laundry_day = misc_small_dfs["superhero_laundry_day"]
alien_abduction_log = misc_small_dfs["alien_abduction_log"]


ID = "idx"

# Simple with one field having a repeated value
three_simple_docs = [
    {ID: 0, "s": "a", "n": 1},
    {ID: 1, "s": "b", "n": 2},
    {ID: 2, "s": "b", "n": 3},
]

# Schema not "perfect": Some missing fields here and there.
# Good to test no-sql cases
nums_and_lans = [
    {ID: 1, "en": "one", "fr": "un", "sp": "uno", "so_far": [1]},
    {ID: 2, "en": "two", "fr": "deux", "so_far": [1, 2]},  # sp missing
    {ID: 3, "en": "three", "fr": "trois", "sp": "tres", "so_far": [1, 2, 3]},
    {
        ID: 4,
        "en": "four",
        "fr": "quatre",
        "sp": "cuatro",
        "so_far": [1, 2, 3, 4],
    },
    {
        ID: 5,
        "en": "five",
        "sp": "cinco",
        "so_far": [1, 2, 3, 4, 5],
    },  # fr missing
]

# Stable schema
# Groupby possibilities (see number: Several unique values) -- this allows to test filtering more naturally
feature_cube = [
    {ID: 1, "number": 6, "color": "red", "dims": {"x": 2, "y": 3}},
    {ID: 2, "number": 6, "color": "blue", "dims": {"x": 3, "y": 2}},
    {ID: 3, "number": 10, "color": "red", "dims": {"x": 2, "y": 5}},
    {ID: 4, "number": 10, "color": "red", "dims": {"x": 5, "y": 2}},
    {ID: 5, "number": 15, "color": "red", "dims": {"x": 3, "y": 5}},
    {ID: 6, "number": 15, "color": "blue", "dims": {"x": 3, "y": 5}},
    {ID: 7, "number": 15, "color": "blue", "dims": {"x": 5, "y": 3}},
]

# Sequence Annotations
# Demoing stream annotations with source and time interval
sequence_annots = [
    {"source": "audio", "bt": 5, "tt": 7, "annot": "cat"},
    {
        "source": "audio",
        "bt": 6,
        "tt": 9,
        "annot": "dog",
        "comments": "barks and chases cat away",
    },
    {"source": "visual", "bt": 5, "tt": 8, "annot": "cat"},
    {
        "source": "visual",
        "bt": 6,
        "tt": 15,
        "annot": "dog",
        "comments": "dog remains in view after bark ceases",
    },
]

####### Rock bands #####################################################################################################

rock_bands = [
    {
        ID: "pink_floyd_id",
        "name": "Pink Floyd",
        "members": [
            {"_id": "1", "firstname": "Roger", "lastname": "Waters"},
            {"_id": "2", "firstname": "Nick", "lastname": "Mason"},
            {"_id": "3", "firstname": "Syd", "lastname": "Barrett"},
            {"_id": "4", "firstname": "Richard", "lastname": "Write"},
        ],
    },
    {
        ID: "the_doors_id",
        "name": "The Doors",
        "members": [
            {"_id": "1", "firstname": "Jim", "lastname": "Morrison"},
            {"_id": "2", "firstname": "Ray", "lastname": "Manzarek"},
            {"_id": "3", "firstname": "Robby", "lastname": "Krieger"},
            {"_id": "4", "firstname": "John", "lastname": "Densmore"},
        ],
    },
]

####### BDFLs ##########################################################################################################
# ubtained with:
# from scrapyng.tables import get_tables_from_url (if no scrapyng, use ut.webscrape.tables)
# t = get_tables_from_url('https://en.wikipedia.org/wiki/Benevolent_dictator_for_life')[0]
# t = t[['Name', 'Project', 'Type']].to_dict(orient='records')

bdfl = [
    {
        "Name": "Sylvain Benner",
        "Project": "Spacemacs",
        "Type": "Community-driven Emacs distribution",
    },
    {
        "Name": "Vitalik Buterin",
        "Project": "Ethereum",
        "Type": "Blockchain-based cryptocurrency",
    },
    {
        "Name": "Dries Buytaert",
        "Project": "Drupal",
        "Type": "Content management framework",
    },
    {
        "Name": "Haoyuan Li",
        "Project": "Alluxio",
        "Type": "Data Orchestration System",
    },
    {
        "Name": "Evan Czaplicki",
        "Project": "Elm",
        "Type": "Front-end web programming language",
    },
    {
        "Name": "David Heinemeier Hansson",
        "Project": "Ruby on Rails",
        "Type": "Web framework",
    },
    {
        "Name": "Rich Hickey",
        "Project": "Clojure",
        "Type": "Programming language",
    },
    {
        "Name": "Adrian Holovaty and Jacob Kaplan-Moss",
        "Project": "Django",
        "Type": "Web framework",
    },
    {
        "Name": "Laurent Destailleur",
        "Project": "Dolibarr ERP CRM",
        "Type": "Software suite for Enterprise Resource Planning and Customer "
        "Relationship Management",
    },
    {
        "Name": "Francois Chollet",
        "Project": "Keras",
        "Type": "Deep learning framework",
    },
    {
        "Name": "Xavier Leroy",
        "Project": "OCaml",
        "Type": "Programming language",
    },
    {
        "Name": "Yukihiro Matsumoto (Matz)",
        "Project": "Ruby",
        "Type": "Programming language",
    },
    {
        "Name": "Wes McKinney",
        "Project": "Pandas",
        "Type": "Python data analysis library",
    },
    {"Name": "Bram Moolenaar", "Project": "Vim", "Type": "Text editor"},
    {
        "Name": "Matt Mullenweg [a]",
        "Project": "WordPress",
        "Type": "Content management framework",
    },
    {
        "Name": "Martin Odersky",
        "Project": "Scala",
        "Type": "Programming language",
    },
    {"Name": "Taylor Otwell", "Project": "Laravel", "Type": "Web framework"},
    {
        "Name": "Theo de Raadt",
        "Project": "OpenBSD",
        "Type": "A Unix-like operating system",
    },
    {
        "Name": "Ton Roosendaal[b]",
        "Project": "Blender",
        "Type": "3D computer graphics software",
    },
    {
        "Name": "Sébastien Ros",
        "Project": "Orchard Project",
        "Type": "Content management system",
    },
    {
        "Name": "Mark Shuttleworth[c]",
        "Project": "Ubuntu",
        "Type": "Linux distribution",
    },
    {"Name": "Don Syme[d]", "Project": "F#", "Type": "Programming language"},
    {
        "Name": "Linus Torvalds[e]",
        "Project": "Linux",
        "Type": "Operating system kernel",
    },
    {
        "Name": "José Valim",
        "Project": "Elixir",
        "Type": "Programming language",
    },
    {
        "Name": "Pauli Virtanen",
        "Project": "SciPy",
        "Type": "Python library used for scientific and technical computing",
    },
    {
        "Name": "Patrick Volkerding",
        "Project": "Slackware",
        "Type": "GNU/Linux distribution",
    },
    {
        "Name": "Nathan Voxland",
        "Project": "Liquibase",
        "Type": "Database schema management",
    },
    {
        "Name": "Shaun Walker",
        "Project": "DotNetNuke",
        "Type": "Web application framework",
    },
    {"Name": "Larry Wall", "Project": "Perl", "Type": "Programming language"},
    {
        "Name": "Jeremy Soller[37]",
        "Project": "Redox",
        "Type": "Operating system",
    },
    {
        "Name": "Eugen Rochko",
        "Project": "Mastodon",
        "Type": "open source, decentralized social network",
    },
    {
        "Name": "Dylan Araps",
        "Project": "KISS Linux",
        "Type": "a bare-bones Linux distribution based on musl libc and BusyBox",
    },
    {
        "Name": "Gavin Mendel-Gleason[f]",
        "Project": "TerminusDB",
        "Type": "Open-source graph database for knowledge graph representation",
    },
]



"""Wrapping tools

A lot of what is defined here are functions that are used to transform data.
More precisely, encode and decode data depending on it's format, file extension, etc.

"""

import os
import re
import pickle
from functools import partial
import io
from io import BytesIO
import pandas as pd
import json
import pickle
from posixpath import splitext
from typing import Mapping, Union, Callable, Dict, TypeVar, KT, VT, BinaryIO

import numpy as np
from dol import Pipe, wrap_kvs, written_bytes, store_decorator
from dol.zipfiledol import file_or_folder_to_zip_file
from i2 import mk_sentinel, Sig

from tabled.util import identity, split_keys, auto_decode_bytes

# TODO: Merge with imbed.base extension-based codec mapping, and move here
# TODO: Add some registry functionality to this?
# TODO: Merge with dol's extension-based codec mapping (routing)
# TODO: Merge with codec-matching ("routing"?) functionalities of dol
# TODO: Move the extension-based codec stuff to tabled, replacing current DfFiles etc.

Obj = TypeVar("Obj")
KeyFunc = Callable[[Obj], KT]
Extension = TypeVar("Extension")
DfDecoder = Callable[[Obj], pd.DataFrame]
dflt_not_found_sentinel = mk_sentinel("dflt_not_found_sentinel")


# --------------------------------------------------------------------------------------
# Helpers for file extensions and protocols


def _get_extension(path: str) -> str:
    return splitext(path)[1]


def _lower_case_without_dot_prefix(ext: str) -> str:
    return ext[1:].lower()


def _key_property(
    key: KT, *, extract_prop=_get_extension, prop_egress=_lower_case_without_dot_prefix
):
    ext = extract_prop(key)
    return prop_egress(ext)


def file_extension(key: str) -> Extension:
    """Get the file extension from a key

    >>> file_extension('hello.world')
    'world'
    >>> file_extension('hello')
    ''
    """
    return _key_property(
        key, extract_prop=_get_extension, prop_egress=_lower_case_without_dot_prefix
    )


get_file_ext = file_extension  # back-compatibility alias


# TODO: Deprecated
def get_extension(key: str) -> str:
    """Return the extension of a file path.

    Note that it includes the dot.

    >>> get_extension('hello.world')  # doctest: +SKIP
    '.world'

    If there's no extension, it returns an empty string.

    >>> get_extension('hello')  # doctest: +SKIP
    ''

    """
    print(f"Deprecating get_extension. Consider using ")
    return _key_property(key, extract_prop=_get_extension, prop_egress=identity)


protocol_re = re.compile(r"([a-zA-Z0-9]+)://")


def get_protocol(url: str):
    """Get the protocol of a url

    >>> get_protocol('https://www.google.com')
    'https'
    >>> get_protocol('file:///home/user/file.txt')
    'file'

    The function returns None if no protocol is found:

    >>> assert get_protocol('no_protocol_here') is None

    """
    m = protocol_re.match(url)
    if m:
        return m.group(1)


def key_func_mapping(
    obj: Obj,
    mapping: Mapping[KT, VT],
    key: KeyFunc = identity,
    not_found_sentinel=dflt_not_found_sentinel,
) -> VT:
    """Map an object to a value based on a key function"""
    return mapping.get(key(obj), not_found_sentinel)


# --------------------------------------------------------------------------------------
# Helpers for the codec mappings


def if_extension_not_present_add_it(filepath, extension):
    if not filepath.endswith(extension):
        return filepath + extension
    return filepath


def if_extension_present_remove_it(filepath, extension):
    if filepath.endswith(extension):
        return filepath[: -len(extension)]
    return filepath


def save_df_to_zipped_tsv(df: pd.DataFrame, name: str, sep="\t", index=False, **kwargs):
    """Save a dataframe to a zipped tsv file."""
    name = if_extension_present_remove_it(name, ".zip")
    name = if_extension_present_remove_it(name, ".tsv")
    tsv_filepath = f"{name}.tsv"
    zip_filepath = f"{tsv_filepath}.zip"
    df.to_csv(tsv_filepath, sep=sep, index=index, **kwargs)

    file_or_folder_to_zip_file(tsv_filepath, zip_filepath)


from i2 import LiteralVal
from tabled.util import is_instance_of
from operator import methodcaller


def map_values(
    func: Callable,
    d: dict,
    *,
    except_condition=is_instance_of(LiteralVal),
    except_handler=methodcaller("__call__"),
):
    """Map values of a dictionary, except for those that satisfy a condition.

    The `except_condition` is a function that takes a value and returns a boolean.
    If the condition is True, the value is not mapped.
    Instead, the `except_handler` is called with the value, and the result is used as
    the new value (often, the value is left unchanged).

    The default `except_condition` is `is_instance_of(LiteralVal)`, which is a function
    that returns True if the value is to be taken litterally.
    The default `except_handler` is `methodcaller('__call__')`, which will extract
    the litteral value from the `LiteralVal` object.

    >>> map_values(lambda x: x * 10, {'a': 1, 'b': LiteralVal(2), 'c': 3})
    {'a': 10, 'b': 2, 'c': 30}

    """
    return {
        k: except_handler(v) if except_condition(v) else func(v) for k, v in d.items()
    }


# --------------------------------------------------------------------------------------
# The codec mappings

# TODO: Idea: set up the extension_codec plugin system, and use it to set up the default
#   codec mappings, keeping encoder and decoder close.
#   Then, we can have a INCLUDE_INDEX_BY_DEFAULT flat that can help us manage the
#   fact that:
#       if index=True in encoder,  decoder needs to include index_col=0
#       if index=False in encoder, decoder needs to include index_col=None

USE_INDEX = True  # set here for the encoders
INDEX_COL = 0 if USE_INDEX else None  # ... and this will be used for the decoders

_extension_to_encoder = split_keys(
    {
        # csv files
        # note: if index=True, decoder needs to include index_col=0
        #       if index=False, decoder needs to include index_col=None
        "csv txt": partial(pd.DataFrame.to_csv, index=USE_INDEX),
        # tab-separated files
        "tsv": partial(
            pd.DataFrame.to_csv,
            index=USE_INDEX,
            sep="\t",
            escapechar="\\",
            quotechar='"',
        ),
        # json files
        "json": pd.DataFrame.to_json,
        # html tables
        "html": partial(pd.DataFrame.to_html, index=USE_INDEX),
        # pickle files
        "p pickle pkl": pd.DataFrame.to_pickle,
        # numpy arrays
        "npy": LiteralVal(written_bytes(np.save, obj_arg_position_in_writer=1)),
        # parquet format
        "parquet": LiteralVal(
            written_bytes(pd.DataFrame.to_parquet, obj_arg_position_in_writer=0)
        ),
        # zip-compressed tsv (custom implementation)
        "zip": LiteralVal(save_df_to_zipped_tsv),
        # feather format
        "feather": pd.DataFrame.to_feather,
        # hdf5 format (Hierarchical Data Format)
        "h5 hdf5": pd.DataFrame.to_hdf,
        # stata files
        "stata dta": partial(pd.DataFrame.to_stata, write_index=USE_INDEX),
        # sql queries
        "sql sqlite": pd.DataFrame.to_sql,
        # Google BigQuery
        "gbq": pd.DataFrame.to_gbq,
        # ------------ extensions requiring extra dependencies ------------
        # excel files
        "xls xlsx": partial(
            pd.DataFrame.to_excel, index=USE_INDEX
        ),  # Need: pip install openpyxl, xlrd
        # xml files
        "xml": pd.DataFrame.to_xml,  # Need: pip install lxml
        # parquet format
        "parquet": pd.DataFrame.to_parquet,  # Need: pip install pyarrow, fastparquet
        # feather format
        "feather": pd.DataFrame.to_feather,  # Need: pip install pyarrow
        # orc format
        "orc": pd.DataFrame.to_orc,  # Need: pip install pyarrow
        # # sas files
        # 'sas': written_bytes(pd.DataFrame.to_sas),  # Need: pip install sas7bdat
        # # SPSS files
        # 'sav': written_bytes(pd.DataFrame.to_spss),  # Need: pip install pyreadstat
    }
)

extension_to_encoder = map_values(written_bytes, _extension_to_encoder)


from dol.util import read_from_bytes

_extension_to_decoder = split_keys(
    {
        # csv and text
        "csv txt": partial(pd.read_csv, index_col=INDEX_COL),
        # tab-separated files
        "tsv": partial(pd.read_csv, sep="\t", index_col=INDEX_COL),
        # parquet format
        "parquet": pd.read_parquet,
        # json format
        "json": partial(pd.read_json, orient="records"),
        # html tables
        "html": partial(pd.read_html, index_col=INDEX_COL),
        # pickle files
        "p pickle pkl": pickle.load,
        # xml files
        "xml": pd.read_xml,
        # sql queries
        "sql sqlite": pd.read_sql,
        # feather format
        "feather": pd.read_feather,
        # stata files
        "stata dta": partial(pd.read_stata, index_col=INDEX_COL),
        # sas files
        "sas": pd.read_sas,
        # extensions requiring extra dependencies
        # hdf5 format (Hierarchical Data Format)
        "h5 hdf5": pd.read_hdf,  # Need: pip install tables
        # excel files
        "xls xlsx": partial(
            pd.read_excel, index_col=INDEX_COL
        ),  # Need: pip install openpyxl, xlrd
        # parquet format
        "parquet": pd.read_parquet,  # Need: pip install pyarrow, fastparquet
        # feather format
        "feather": pd.read_feather,  # Need: pip install pyarrow
        # orc format
        "orc": pd.read_orc,  # Need: pip install pyarrow
        # SPSS files
        "sav": pd.read_spss,  # Need: pip install pyreadstat
    }
)

extension_to_decoder = map_values(read_from_bytes, _extension_to_decoder)

dflt_ext_mapping = extension_to_decoder  # back-compatibility alias


# --------------------------------------------------------------------------------------
# Operations on the codec mappings


def get_codec_mappings(
    *,
    extension_to_encoder=extension_to_encoder,
    extension_to_decoder=extension_to_decoder,
):
    return dict(
        encoders=extension_to_encoder,
        decoders=extension_to_decoder,
    )


def print_current_mappings():
    from pprint import pprint

    pprint("Current encoder and decoder mappings:")
    pprint(get_codec_mappings())


def add_extension_codec(extension=None, *, encoder=None, decoder=None):
    """
    Add an extension-based encoder and decoder to the extension-code mapping.

    Sure, you could just edit the underlying dictionaries directly, but the design gods
    would not be pleased.

    If no arguments are passed, it will print the current mappings.

    Returns: None (it just adds the in-memory mappings)

    """
    if extension is None and encoder is None and decoder is None:
        # Print the current mappings
        # (The design gods would hate this, for sure)
        return print_current_mappings()

    if encoder is not None:
        extension_to_encoder[extension] = encoder
    if decoder is not None:
        extension_to_decoder[extension] = decoder


# --------------------------------------------------------------------------------------
# Main facades to the codec mappings


def extension_based_decoding(k, v, *, extension_to_decoder=extension_to_decoder):
    """Decode a value based on the extension of the key."""
    ext = file_extension(k)
    decoder = extension_to_decoder.get(ext, None)
    if decoder is None:
        suffix_msg = (
            f"This happened with key: {k}\n"
            "Note that you can add your own extension-based decoder to resolve this."
        )
        if ext != "":
            raise ValueError(
                f"Your DECODER doesn't support this extension: {ext}\n{suffix_msg}"
            )
        else:
            raise ValueError(
                f"Your DECODER doesn't support empty extensions.\n{suffix_msg}"
            )
    return decoder(v)


def extension_based_encoding(k, v, *, extension_to_encoder=extension_to_encoder):
    """Encode a value based on the extension of the key."""
    ext = file_extension(k)
    encoder = extension_to_encoder.get(ext, None)
    if encoder is None:
        suffix_msg = (
            f"This happened with key: {k}\n"
            "Note that you can add your own extension-based encoder to resolve this."
        )
        if ext != "":
            raise ValueError(
                f"Your ENCODER doesn't support this extension: {ext}\n{suffix_msg}"
            )
        else:
            raise ValueError(
                f"Your ENCODER doesn't support empty extensions.\n{suffix_msg}"
            )
    return encoder(v)


@store_decorator
def extension_based_wrap(store=None):
    """Add extension-based encoding and decoding to a store."""
    return wrap_kvs(
        store,
        postget=extension_based_decoding,
        preset=extension_based_encoding,
    )


# ----------------------------------------
# moved from tabled.base


def resolve_to_dataframe(
    data,
    ext: Extension,
    ext_mapping: Mapping = extension_to_decoder,
    **extra_decoder_kwargs,
) -> pd.DataFrame:
    """Get a dataframe from a (data, ext) pair"""
    if ext.startswith("."):
        ext = ext[1:]
    decoder = key_func_mapping(
        ext,
        ext_mapping,
        key=identity,
        not_found_sentinel=None,  # TODO
    )
    if decoder is not None:
        # pluck out any key-value pairs of extra_decoder_kwargs whose names are
        # arguments of decoder. (This is needed since extra_decoder_kwargs can be
        # servicing multiple decoders, and we don't want to pass arguments to the
        # wrong decoder.)
        extra_decoder_kwargs = Sig(decoder).map_arguments(
            (), extra_decoder_kwargs, allow_partial=True, allow_excess=True
        )
        # decode the data
        return decoder(data, **extra_decoder_kwargs)
    else:
        raise ValueError(f"Don't know how to handle extension: {ext}")


df_from_data_given_ext = resolve_to_dataframe  # back-compatibility alias


def df_from_data_according_to_key(
    data: Obj, mapping: Dict[Extension, DfDecoder], key: KT, **extra_decoder_kwargs
):
    """Get a dataframe from a (data, mapping, key) triple"""
    decoder = key_func_mapping(data, mapping, key=key, not_found_sentinel=None)
    if decoder is None:
        raise ValueError(f"Don't know how to handle key: {key}")
    return decoder(data, **extra_decoder_kwargs)


# TODO: Implemented a plugin system (routing) for io resolution concern in disucssion #8
#   See https://github.com/i2mint/tabled/discussions/8#discussion-7519188
#   Just don't know if it's worth the complexity yet.
#   If you find yourself editing default_io_resolver or specifying a lot of custom ones,
#   it might be time to use the plugin system instead.

TableSrc = Union[str, bytes, BinaryIO]
BinaryIOCaster = Callable[[TableSrc], BinaryIO]


# TODO: Routing pattern!
def default_io_resolver(src: TableSrc) -> BinaryIO:
    if isinstance(src, str):
        if os.path.isfile(src):
            return open(src, "rb")
        elif get_protocol(src) in {"http", "https"}:
            import requests

            return BytesIO(requests.get(src).content)
        elif get_protocol(src) == "graze":
            from graze import graze  # pip install graze

            _, url = src.split("://", 1)
            return BytesIO(graze(url))
        else:
            raise ValueError(f"Can't handle source: {src}")
    elif isinstance(src, bytes):
        return BytesIO(src)
    elif hasattr(src, "read"):
        return src
    else:
        raise ValueError(f"Can't handle source: {src}")



"""Based functionality for tabled"""

# TODO: Completely change so that
#   * routing is more general (not only conditioned on ext, but function of key,
#   and possibly data
#   * Allow on-the-fly parametrization (e.g. sep='\t', index_col=False...)
from functools import partial
from io import BytesIO
from typing import (
    Mapping,
    KT,
    VT,
    Dict,
    Iterable,
    Union,
)

import pandas as pd

from dol import Files, KvReader

from tabled.util import identity, split_keys


from tabled.wrappers import (
    TableSrc,
    KeyFunc,
    Extension,
    DfDecoder,
    key_func_mapping,  # Map an object to a value based on a key function
    dflt_ext_mapping,  # Default extension mapping for various file types (DEPRECATED)
    resolve_to_dataframe,  # Get a dataframe from a (data, ext) pair
    df_from_data_given_ext,  # Alias for resolve_to_dataframe for backward compatibility
    df_from_data_according_to_key,  # Get a dataframe from a (data, mapping, key) triple
    file_extension,  # Get the file extension from a key
    get_protocol,  # Get the protocol of a URL
    default_io_resolver,  # Default IO resolver for various sources
)


# Obj = TypeVar('Obj')
# KeyFunc = Callable[[Obj], KT]
# Extension = TypeVar('Extension')
# DfDecoder = Callable[[Obj], pd.DataFrame]
# dflt_not_found_sentinel = mk_sentinel('dflt_not_found_sentinel')


# def key_func_mapping(
#     obj: Obj,
#     mapping: Mapping[KT, VT],
#     key: KeyFunc = identity,
#     not_found_sentinel=dflt_not_found_sentinel,
# ) -> VT:
#     """Map an object to a value based on a key function"""
#     return mapping.get(key(obj), not_found_sentinel)


class KeyFuncReader(KvReader):
    def __init__(self, mapping: Mapping[KT, VT], key: KeyFunc = identity):
        self.mapping = mapping
        self.key = key

    def __getitem__(self, k):
        return self.mapping[self.key(k)]

    def __contains__(self, k):
        return self.key(k) in self.mapping

    def __iter__(self):
        return iter(self.mapping)

    def __len__(self):
        return len(self.mapping)

    def __repr__(self):
        return f"{type(self).__name__}({self.mapping}, key={self.key})"


def convert_collection_to_dataframe_if_possible(x):
    if isinstance(x, pd.DataFrame):
        return x
    elif isinstance(x, (dict, list, tuple)):
        return pd.DataFrame(x)
    elif isinstance(x, (pd.Series, pd.Index, pd.MultiIndex)):
        return x.to_frame()
    elif isinstance(x, pd.Index):
        return x.to_frame()
    else:
        return x


# TODO: Handle the zip case more cleanly, and generalize to other double-extensions
def get_table(
    table_src: TableSrc = None,
    *,
    ext=None,
    ext_mapping=dflt_ext_mapping,
    resolve_to_io=default_io_resolver,
    **extra_decoder_kwargs,
) -> pd.DataFrame:
    """
    Get a table from a variety of sources.
    """
    # If table_src is None, the user is trying to fix the parameters of the function
    if table_src is None:
        return partial(
            get_table,
            ext=ext,
            ext_mapping=ext_mapping,
            resolve_to_io=resolve_to_io,
            **extra_decoder_kwargs,
        )

    table_src = convert_collection_to_dataframe_if_possible(table_src)
    if isinstance(table_src, pd.DataFrame):
        return table_src

    if ext is None and isinstance(table_src, str):
        key = table_src
        ext = file_extension(key)
        if ext == "zip":
            # compute the next extension
            next_ext = file_extension(key[: -len(".zip")])
            if next_ext:
                ext = f"{next_ext}.{ext}"  # e.g. 'csv.zip'

    # TODO: Here's a great waste, since many of our table reading functions can
    #       take file-like (paths, io objects) as input. Should make wrappers.py so that
    #       one decoder can be given, and a input-type aware function can transform
    #       the input to the right type for that particular decoder
    # convert to bytes if not so already
    if not isinstance(table_src, bytes):
        io_reader = resolve_to_io(table_src)
        table_src = io_reader.read()

    if ext.endswith(".zip"):
        from dol import zip_decompress

        table_src = zip_decompress(table_src)  # get decompressed bytes
        ext = ext[: -len(".zip")]  # remove the '.zip' from the extension

    return resolve_to_dataframe(
        table_src,
        ext=ext,
        ext_mapping=ext_mapping,
        **extra_decoder_kwargs,
    )


from tabled.wrappers import extension_based_encoding, extension_to_encoder


# TODO: Make the logic independent from local files assumption.
# TODO: Add filtering functionality in init? (By function, regex, extension?)
class DfFiles(Files):
    """A key-value store providing values as pandas.DataFrames.

    Use Case: You have a bunch of files in a folder, all corresponding to some
    dataframes that were saved in some way. You want to a key-value store whose values
    are the (decoded) dataframes corresponding to the files in the folder.

    Args:
    rootdir: A root directory.
    extension_decoder_mapping: A mapping from file extensions to functions that can
        read the dataframes
    extra_decoder_kwargs: Extra arguments to pass to the decoder functions.

    """

    def __init__(
        self,
        rootdir: str,
        *,
        extension_encoder_mapping: Dict[Extension, DfDecoder] = extension_to_encoder,
        extension_decoder_mapping: Dict[Extension, DfDecoder] = dflt_ext_mapping,
        extra_encoder_kwargs: Union[dict, Iterable] = (),
        extra_decoder_kwargs: Union[dict, Iterable] = (),
    ):
        super().__init__(rootdir)

        self.key_to_ext = file_extension

        self.extension_encoder_mapping = extension_encoder_mapping
        self.extension_decoder_mapping = extension_decoder_mapping

        extra_encoder_kwargs = dict(extra_encoder_kwargs)
        extra_decoder_kwargs = dict(extra_decoder_kwargs)

        self.bytes_and_ext_to_df = partial(
            df_from_data_given_ext,
            ext_mapping=extension_decoder_mapping,
            **extra_decoder_kwargs,
        )

        # self.extension_based_encoder = partial(
        #     extension_based_encoding, extension_encoder_mapping
        # )

    def __getitem__(self, k):
        ext = self.key_to_ext(k)
        # data = BytesIO(super().__getitem__(k))
        data = super().__getitem__(k)
        return self.bytes_and_ext_to_df(data, ext)

    # TODO: Implement this. This works:
    # from tabled import extension_base_wrap
    # extension_base_wrap(Files)
    # But not this:
    def __setitem__(self, k, v):
        bytes_ = extension_based_encoding(
            k, v, extension_to_encoder=self.extension_encoder_mapping
        )
        # bytes_ = self.extension_based_encoding(v, file_extension(k))
        return super().__setitem__(k, bytes_)

    def __delitem__(self, k):
        return super().__delitem__(k)


class DfReader(DfFiles):
    def __setitem__(self, k, v):
        raise NotImplementedError("DfReader is a read-only store.")

    def __delitem__(self, k):
        raise NotImplementedError("DfReader is a read-only store.")


DfLocalFileReader = DfReader  # back-compatibility: TODO: Issue warning on use


# ------------------------------------------------------------------------------
# Wrapping a DataFrame to provide a key-value interface

import pandas as pd
from collections.abc import Mapping


def validate_fields(df, key_fields, value_columns):
    all_columns = df.columns.tolist()
    all_index_levels = list(df.index.names)

    for field in key_fields:
        if field not in all_columns and field not in all_index_levels:
            raise ValueError(
                f"Field {field} not found in the DataFrame columns or index levels."
            )

    for col in value_columns:
        if col not in all_columns:
            raise ValueError(f"Column {col} not found in the DataFrame.")


# TODO: Does it accommodate the case where I want to define my key_fields to be the
# UNNAMED index of the dataframe, plus a column.
# For example, consider (and add to the test) the case where I have df.set_index('A'),
# but remove the name "A" from the index. Now I can't say key_fields=['A', 'B'] anymore.
class DataframeKvReader(Mapping):
    """
    A class to wrap a DataFrame and provide a Mapping interface where keys are
    combinations of specified columns or index levels, and values are sub-dataframes of specified columns.

    Example usage:
    >>> df = pd.DataFrame({
    ...     'A': [1, 2, 1],
    ...     'B': [4, 5, 4],
    ...     'C': [7, 8, 9],
    ...     'D': [10, 11, 12]
    ... })
    >>> df
       A  B  C   D
    0  1  4  7  10
    1  2  5  8  11
    2  1  4  9  12
    >>> kv_reader = DataframeKvReader(df, ['A', 'B'], ['C', 'D'])
    >>> key = (1, 4)
    >>> kv_reader[key].reset_index(drop=True)  # doctest: +NORMALIZE_WHITESPACE
       C   D
    0  7  10
    1  9  12
    >>> list(kv_reader) == [(1, 4), (2, 5)]
    True

    But what if one (or more) of the key fields is an index level?
    The DataframeKvReader can handle that too:

    >>> df = df.set_index(['A'])
    >>> df  # doctest: +NORMALIZE_WHITESPACE
       B  C   D
    A
    1  4  7  10
    2  5  8  11
    1  4  9  12
    >>> kv_reader = DataframeKvReader(df, ['A', 'B'], ['C', 'D'])
    >>> key = (1, 4)
    >>> kv_reader[key].reset_index(drop=True)  # doctest: +NORMALIZE_WHITESPACE
       C   D
    0  7  10
    1  9  12
    >>> list(kv_reader) == [(1, 4), (2, 5)]
    True

    """

    def __init__(self, df, key_fields, value_columns=None):
        """
        Initialize the DataframeKvReader.

        Parameters:
        df (pd.DataFrame): The DataFrame to wrap.
        key_fields (str or list of str): Fields (columns or index levels) to use as keys.
        value_columns (str or list of str): Column(s) to use as values. Default is all columns.

        """
        if value_columns is None:
            value_columns = df.columns.tolist()
        validate_fields(df, key_fields, value_columns)
        self.key_fields = key_fields if isinstance(key_fields, list) else [key_fields]
        self.value_columns = (
            value_columns if isinstance(value_columns, list) else [value_columns]
        )

        # Check if key fields are all index levels
        self.is_index_key = all(field in df.index.names for field in self.key_fields)

        if not self.is_index_key:
            # If key fields are not all index levels, reset index and set new index
            self.df = (
                df.reset_index().set_index(self.key_fields, drop=False).sort_index()
            )
        else:
            # If key fields are all index levels, ensure the DataFrame is sorted by those key fields
            self.df = df.sort_index()

    def __getitem__(self, key):
        if not isinstance(key, tuple):
            key = (key,)
        try:
            sub_df = self.df.loc[key, self.value_columns]
            if isinstance(sub_df, pd.Series):
                sub_df = sub_df.to_frame().T
            return sub_df
        except KeyError:
            raise KeyError(f"Key {key} not found")

    def __iter__(self):
        keys = self.df.index.drop_duplicates().tolist()
        return iter(keys)

    def __len__(self):
        return self.df.index.nunique()

    def __contains__(self, key):
        return key in self.df.index

    def __repr__(self):
        return f"{type(self).__name__}(df=<{len(self.df)} rows>, key_fields={self.key_fields}, value_columns={self.value_columns})"



import pandas as pd
from tabled.base import DataframeKvReader


# Test cases
def test_DataframeKvReader():

    df = pd.DataFrame(
        {"A": [1, 2, 1], "B": [4, 5, 4], "C": [7, 8, 9], "D": [10, 11, 12]}
    )

    kv_readers = [
        DataframeKvReader(df, ["A", "B"], ["C", "D"]),
        DataframeKvReader(df.set_index(["A", "B"]), ["A", "B"], ["C", "D"]),
        DataframeKvReader(df.set_index("A"), ["A", "B"], ["C", "D"]),
    ]

    for kv_reader in kv_readers:
        # Accessing values
        key = (1, 4)
        expected_df = pd.DataFrame([{"C": 7, "D": 10}, {"C": 9, "D": 12}], index=[0, 2])
        assert (
            kv_reader[key]
            .reset_index(drop=True)
            .equals(expected_df.reset_index(drop=True))
        )
        assert list(kv_reader) == [(1, 4), (2, 5)]


import os
import shutil
import tempfile
import pandas as pd
import pytest
from tabled import DfFiles


def test_df_files_functionality():
    """Test DfFiles functionality matching the demo notebook"""

    # Test data dictionary - same as notebook
    misc_small_dicts = {
        "fantasy_tavern_menu": {
            "item": ["Dragon Ale", "Elf Bread", "Goblin Stew"],
            "price": [7.5, 3.0, 5.5],
            "is_alcoholic": [True, False, False],
            "servings_left": [12, 25, 8],
        },
        "alien_abduction_log": {
            "abductee_name": ["Bob", "Alice", "Zork"],
            "location": ["Kansas City", "Roswell", "Jupiter"],
            "duration_minutes": [15, 120, 30],
            "was_returned": [True, False, True],
        },
    }

    # Create temporary directory for testing
    rootdir = os.path.join(tempfile.gettempdir(), "tabled_df_files_test")
    if os.path.exists(rootdir):
        shutil.rmtree(rootdir)
    os.makedirs(rootdir)

    # Initialize DfFiles
    df_files = DfFiles(rootdir)

    # Verify it starts empty
    assert list(df_files) == []

    # Create DataFrames from test data
    fantasy_tavern_menu_df = pd.DataFrame(misc_small_dicts["fantasy_tavern_menu"])
    alien_abduction_log_df = pd.DataFrame(misc_small_dicts["alien_abduction_log"])

    # Save DataFrames using different formats
    df_files["fantasy_tavern_menu.csv"] = fantasy_tavern_menu_df
    df_files["alien_abduction_log.json"] = alien_abduction_log_df

    # Read data back and verify
    saved_df = df_files["fantasy_tavern_menu.csv"]
    pd.testing.assert_frame_equal(saved_df, fantasy_tavern_menu_df)

    # Test MutableMapping interface
    assert len(df_files) == 2
    assert sorted(list(df_files)) == [
        "alien_abduction_log.json",
        "fantasy_tavern_menu.csv",
    ]
    assert "fantasy_tavern_menu.csv" in df_files

    # Verify supported extensions
    encoder_extensions = list(df_files.extension_encoder_mapping)
    decoder_extensions = list(df_files.extension_decoder_mapping)

    # Test common formats
    extensions_supported_by_both = sorted(
        set(encoder_extensions) & set(decoder_extensions)
    )
    #      ... Reproduced here since not all the formats are actually working yet

    test_extensions = [
        "csv",
        "feather",  # requires `pip install pyarrow``
        "json",
        "orc",  # requires `pip install pyarrow`` or `pip install fastparquet`
        "parquet",  # requires `pip install pyarrow`` or `pip install fastparquet`
        "pkl",
        "tsv",
        # 'dta',  # TODO: fix
        # 'h5',  # TODO: fix
        # 'html',  # TODO: fix
        # 'sql',  # TODO: fix
        # 'xml',  # TODO: fix
    ]
    for ext in test_extensions:
        filename = f"test_file.{ext}"
        df_files[filename] = fantasy_tavern_menu_df
        df_loaded = df_files[filename]

        # Compare DataFrames ignoring index
        pd.testing.assert_frame_equal(
            fantasy_tavern_menu_df.reset_index(drop=True),
            df_loaded.reset_index(drop=True),
        )

    # Cleanup
    shutil.rmtree(rootdir)



"""To work with html"""

from typing import Callable, Union, Mapping, Optional, Sequence
import re
import io

import pandas as pd

DFLT_CHROME_WAIT = 3


def url_to_html_func(kind="requests") -> Callable:
    """Get a url_to_html function of a given kind."""

    # If kind is a tuple, the first element is the kind and the second element is the kwargs
    # to be used to parametrize the function
    # NOTE: For now, I'm just simply passing the kwargs to the place I think is most
    #  needed, but we can use a more sophisticated way to pass the kwargs to the right
    #  place using i2.call_forgivingly or i2.Sig.extract_args_and_kwargs or such
    if not isinstance(kind, str):
        kind_tuple = kind
        kind, kind_kwargs = kind_tuple
    else:
        kind_kwargs = {}

    url_to_html = None
    if kind == "requests":
        import requests  # pip install requests

        def url_to_html(url):
            r = requests.get(url, **kind_kwargs)
            if r.status_code != 200:
                print(
                    f"An error occured. Returning the response object for you to analyze: {r}"
                )
                return r
            return r.content

    elif kind == "chrome" or kind == "selenium":

        from selenium import webdriver  # pip install selenium
        from time import sleep

        dflt_wait = kind_kwargs.get("wait", DFLT_CHROME_WAIT)

        def url_to_html(url, wait=dflt_wait):
            b = webdriver.Chrome()
            b.get(url)
            if isinstance(wait, (int, float)):
                sleep(wait)
            html = b.page_source
            b.close()
            return html

    else:
        raise ValueError(f"Unknown url_to_html value: {url_to_html}")
    assert callable(url_to_html), "Couldn't make a url_to_html function"

    return url_to_html


get_tables_from_html = pd.read_html


TableFilter = Union[str, Sequence[str], Callable]


def _ensure_table_filter(filt: TableFilter) -> Callable:
    """
    Ensure that the filter is a callable that takes a dataframe and returns a boolean.

    If filt is a string, the column names must contain at least one column that matches filt regex.
    If filt is a list of strings, the column names must contain at least one column that matches
    the regexes in the filt list, for each regex in the list.
    If filt is a callable, it is used as is.

    Raises a ValueError if the filter type is unknown.

    Examples:

    >>> filt_func = _ensure_table_filter('foo')
    >>> bool(filt_func(pd.DataFrame({'foo': [1, 2]})))
    True
    >>> bool(filt_func(pd.DataFrame({'bar': [1, 2]})))
    False
    >>> filt_func = _ensure_table_filter(['foo', 'bar'])
    >>> bool(filt_func(pd.DataFrame({'football': [1, 2], 'baring': [3, 4]})))
    True
    >>> bool(filt_func(pd.DataFrame({'football': [1, 2], 'neither': [3, 4]})))
    False

    But if a same column name matches both regexes, it should return True:

    >>> bool(filt_func(pd.DataFrame({'foobar': [1, 2], 'huh': [3, 4]})))
    True

    """
    if filt is None:
        return lambda x: True
    if isinstance(filt, str):
        # The column names must contain at least one column that matches filt regex
        filt_regex = re.compile(filt)
        return lambda x: x.columns.str.contains(filt_regex).any()
    if isinstance(filt, Sequence):
        # The column names must contain at least one column that matches the
        # regexes in the filt list, for each regex in the list
        filt_regexes = [re.compile(f) for f in filt]
        return lambda x: all(x.columns.str.contains(f).any() for f in filt_regexes)
    if callable(filt):
        return filt
    raise ValueError(f"Unknown filter type: {filt}")


def get_tables_from_url(
    url,
    *,
    url_to_html: Union[Callable, str] = "requests",
    filt: Optional[TableFilter] = None,
    **tables_from_html_kwargs,
):
    """Get's a list of pandas dataframes from tables scraped from a url.
    Note that this will only work with static pages. If the html needs to be rendered dynamically,
    you'll have to get your needed html otherwise (like with selenium).

    >>> url = 'https://en.wikipedia.org/wiki/List_of_musical_instruments'
    >>> tables = get_tables_from_url(url)  # doctest: +SKIP

    If you install selenium and download a chromedriver,
    you can even use your browser to render dynamic html.
    Say, to get updated coronavirus stats without a need to figure out the API
    (I mean, why have to figure out the language of an API, when someone already did that
    for you in their webpage!!):

    ```python
    url = 'https://www.worldometers.info/coronavirus/?utm_campaign=homeAdvegas1?'
    tables = get_tables_from_url(url, url_to_html='chrome')  # doctest: +SKIP
    ```

    To make selenium work:

    - `pip install selenium`
    - Download seleniumdriver here: https://chromedriver.chromium.org/
    - Uzip and put in a place that's on you PATH (run command `echo $PATH` for a list of those places)

    """
    if not callable(url_to_html):
        url_to_html = url_to_html_func(url_to_html)

    filt_func = _ensure_table_filter(filt)

    try:
        html = url_to_html(url)
        tables = get_tables_from_html(io.StringIO(html), **tables_from_html_kwargs)
        return list(filter(filt_func, tables))
    except ValueError as e:
        if len(e.args) > 0:
            msg, *_ = e.args
            if "No tables found" in msg:
                return []
        raise


HTML_TEMPLATE1 = """
<html>
<head>
<style>
  h2 {
    text-align: center;
    font-family: Helvetica, Arial, sans-serif;
  }
  table { 
    margin-left: auto;
    margin-right: auto;
  }
  table, th, td {
    border: 1px solid black;
    border-collapse: collapse;
  }
  th, td {
    padding: 5px;
    text-align: center;
    font-family: Helvetica, Arial, sans-serif;
    font-size: 90%;
  }
  table tbody tr:hover {
    background-color: #dddddd;
  }
  .wide {
    width: 90%; 
  }
</style>
</head>
<body>
"""

HTML_TEMPLATE2 = """
</body>
</html>
"""


def df_to_html(df, title=None):
    ht = ""
    if title is not None:
        ht += f"<h2> {title} </h2>\n"
    ht += df.to_html(classes="wide", escape=False)
    return ht


def df_store_to_html(df_store, sep="\n<br>\n"):
    ht = ""
    for k, df in df_store.items():
        title = re.match(r"[^\d]+", k).group(0)
        ht += df_to_html(df, title)
        ht += sep
    return ht


def dfs_to_html_pretty(dfs, title=None):
    """
    Write an entire dataframe to an HTML file
    with nice formatting.
    Thanks to @stackoverflowuser2010 for the
    pretty printer see https://stackoverflow.com/a/47723330/362951
    """

    if isinstance(dfs, pd.DataFrame):
        ht = df_to_html(dfs, title=title)
    elif isinstance(dfs, Mapping):
        ht = df_store_to_html(dfs)
    else:
        ht = df_store_to_html(dict(enumerate(dfs)))

    return HTML_TEMPLATE1 + ht + HTML_TEMPLATE2


def dfs_to_pdf_bytes(dfs, title=None):
    import weasyprint

    html = dfs_to_html_pretty(dfs, title)
    return weasyprint.HTML(string=html).write_pdf()



"""
Miscellaneous utility functions for tables.
"""

from typing import List, Tuple
import pandas as pd


def scalar_columns(df: pd.DataFrame) -> List:
    """
    Returns the list of columns that are scalar (therefore serializable to CSV)

    More precisely, this function returns the list of columns that contain only
    scalar values (e.g., int, float, str, bool, etc.) and can be saved to a CSV
    file.

    Example:
    >>> import pandas as pd
    >>> df = pd.DataFrame({
    ...     'A': [1, 2, 3],
    ...     'B': ['x', 'y', 'z'],
    ...     'C': [{'a': 1}, {'b': 2}, {'c': 3}],  # Non-serializable column
    ...     'D': [[1, 2], [3, 4], [5, 6]]         # Non-serializable column
    ... })
    >>> scalar_columns(df)
    ['A', 'B']
    """
    import pandas.api.types as pdt

    return [col for col in df.columns if df[col].apply(pdt.is_scalar).all()]
